/*
 jGuard is a security framework based on top of jaas (java authentication and authorization security).
 it is written for web applications, to resolve simply, access control problems.
 version $Name$
 http://sourceforge.net/projects/jguard/

 Copyright (C) 2004  Charles GAY

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 jGuard project home page:
 http://sourceforge.net/projects/jguard/

 */
package net.sf.jguard.ext.authentication.loginmodules;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.security.auth.Subject;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;

import net.sf.jguard.core.authentication.manager.AuthenticationManager;
import net.sf.jguard.core.authentication.manager.AuthenticationManagerFactory;

import net.sf.jguard.ext.authentication.manager.JdbcAuthenticationManager;
import net.sf.jguard.ext.database.ConnectionFactory;

/**
 * LoginModule dedicated to Databases.
 * @author <a href="mailto:diabolo512@users.sourceforge.net">Charles Gay</a>
 * @see LoginModule
 */
public class JdbcLoginModule extends UserLoginModule implements LoginModule {

	static public final String CRED_VALUE = "cred_value";
	static public final String CRED_NAME = "cred_name";
	static public final String NAME = "name";

	private static final String ID = "id";
	private static final Logger logger = Logger.getLogger(JdbcLoginModule.class.getName());

	private int userId = -1;

	private ConnectionFactory connectionFactory;

	private Properties props;
	
	private JdbcAuthenticationManager jdbcAuthManager = null;
	
	/**
	 * initialize database-related loginModule specifying 'digestAlgorithm', tables names,'applicationName', and connection.
         * it can <strong>only run</strong> if a JdbcAuthenticaitonManager is also configured.
	 * parameters.
	 * @param subj
	 * @param cbkHandler
	 * @param sState
	 * @param opts
         */
	public void initialize(Subject subj, CallbackHandler cbkHandler, Map sState, Map opts) {
		super.initialize(subj, cbkHandler, sState, opts);
                AuthenticationManager authenticationManager = AuthenticationManagerFactory.getAuthenticationManager();
                if (authenticationManager instanceof JdbcAuthenticationManager){
                    JdbcAuthenticationManager authManager = (JdbcAuthenticationManager) AuthenticationManagerFactory.getAuthenticationManager();    
                    props = authManager.getProperties();
                    connectionFactory = authManager.getConnectionFactory();
                    jdbcAuthManager = authManager;	
                }else{
                    throw new RuntimeException(" JdbcLoginModule can only be used with a configured JdbcAuthenticationManager ");
                }
		
		
	}

	/**
	 * verify either user is registered or not.
	 *
	 * @see javax.security.auth.spi.LoginModule#login()
	 */
	public boolean login() throws LoginException {
		super.login();
		userId = getUserID(login);
		if (!skipPasswordCheck) {
			boolean passwordValidated = false;
			try{
				passwordValidated = validatePassword(new String(password));
			}catch(LoginException e){
				loginOK = false;
				throw e;
			}
			logger.finest("password validation =" + passwordValidated + " for userId=" + userId);
		}
		return true;

	}

	private boolean validatePassword(String cryptedPassword) throws LoginException {
		boolean authenticated;
		Connection conn2 = null;
		PreparedStatement pst2 = null;
		ResultSet rs2 = null;
		try {
			conn2 = connectionFactory.getConnection();
			pst2 = conn2.prepareStatement((String) props.get("USER_PASSWORD_EXIST"));
			pst2.setInt(1, userId);
			pst2.setString(2, new String(cryptedPassword));
			rs2 = pst2.executeQuery();
			authenticated = rs2.next();
			if (authenticated == false) {
				if (logger.isLoggable(Level.FINEST)) {
					logger.finest(" password does not match ");
				}
				throw new FailedLoginException("login.password.does.not.match");
			}
		} catch (SQLException e2) {
			if (logger.isLoggable(Level.SEVERE)) {
				logger.severe("login() -  connection in " + this.getClass().getName() + " failed " + e2.getMessage());
				logger.severe("USER_PASSWORD_EXIST=" + props.get("USER_PASSWORD_EXIST"));
			}
			throw new LoginException(e2.getMessage());

		} finally {
			try {
				rs2.close();
				pst2.close();
				conn2.close();
			} catch (SQLException e2) {
				if (logger.isLoggable(Level.SEVERE)) {
					logger.severe("login() -  connection in MySQLLoginModule failed " + e2.getMessage());
				}
			}
		}
		return authenticated;
	}

	private int getUserID(String login) throws LoginException {
		boolean authenticated;
		// check the login and grab the user id
		ResultSet rs = null;
		PreparedStatement pst = null;
		Connection conn = null;
		int userId = -1;
		try {
			conn = connectionFactory.getConnection();
			pst = conn.prepareStatement((String) props.get("USER_LOGIN_EXIST"));
			pst.setString(1, login);
			rs = pst.executeQuery();
			authenticated = rs.next();

			if (authenticated == false) {
				if (logger.isLoggable(Level.FINEST)) {
					logger.finest(" user " + login + " does not exists ");
				}
				throw new FailedLoginException("login.user.does.not.exist");
			}

			userId = rs.getInt(JdbcLoginModule.ID);

		} catch (SQLException e1) {
			if (logger.isLoggable(Level.SEVERE)) {
				logger.severe("login() -  connection " + e1.getMessage());
			}
			throw new LoginException(e1.getMessage());
		} finally {
			try {
				rs.close();
				pst.close();
				conn.close();
			} catch (SQLException e2) {
				if (logger.isLoggable(Level.SEVERE)) {
					logger.severe("login() -  connection  failed " + e2.getMessage());
				}
			}
		}

		return userId;
	}

	/**
	 * add Principals and Public/Private credentials to Subject.
	 *
	 * @see javax.security.auth.spi.LoginModule#commit()
	 */
	public boolean commit() throws LoginException {
		if(!loginOK){
			return false;
		}
                Subject subj = null;
		try {
			subj = JdbcAuthenticationManager.getUser(userId);
                } catch (SQLException e) {
			logger.log(Level.SEVERE, "commit()", e);
			throw new LoginException(e.getMessage());
		}
			// we copy the content into the real Subject
			subject.getPrincipals().addAll(subj.getPrincipals());
			subject.getPublicCredentials().addAll(subj.getPublicCredentials());
			subject.getPrivateCredentials().addAll(subj.getPrivateCredentials());
			if (logger.isLoggable(Level.FINEST)) {
				logger.finest(" user authenticated subject=" + subject);
			}
			

		return true;
	}

}
