/*
jGuard is a security framework based on top of jaas (java authentication and authorization security).
it is written for web applications, to resolve simply, access control problems.
version $Name$
http://sourceforge.net/projects/jguard/

Copyright (C) 2004  Charles GAY

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


jGuard project home page:
http://sourceforge.net/projects/jguard/

*/
package net.sf.jguard.ext.authentication.loginmodules;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.security.auth.Subject;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;

import net.sf.jguard.core.CoreConstants;
import net.sf.jguard.core.authentication.AuthenticationException;
import net.sf.jguard.core.authentication.credentials.JGuardCredential;
import net.sf.jguard.core.authentication.manager.AuthenticationManagerFactory;
import net.sf.jguard.ext.authentication.manager.XmlAuthenticationManager;


/**
 * LoginModule configured by the <i>jGuardUsersPrincipals</i> XML file.<br>
 * In webapp environement using JGuardConfiguration, the AuthenticationManager related to the LoginModule is created by AccessFilter.<br>
 * In non-JGuardConfiguration environement, the LoginModule must create its AuthenticationManager, and applicationName
 * is required for this creation.<br>
 * In order to retreive the application name, XmlLoginModule uses the following ways :
 * <ul>
 * <li>trough vm arg : <code>net.sf.jguard.application.name</code> VM arg</li>
 * <li>trough vm arg : <code>com.sun.management.jmxremote.login.config</code>
 *  if you have already defined this property because you use JMX.
 *  Do not set application name through this property if you are not using JMX !</li>
 * </ul>
 * If no applicationName is explicitly passed to the application, default application name "other" is used.
 * @author <a href="mailto:diabolo512@users.sourceforge.net">Charles Gay</a>
 * @see LoginModule
 */
public class XmlLoginModule extends UserLoginModule implements LoginModule{


	private static final String LOGIN_ERROR = "login.error";
	private static final String PASSWORD = "password";
	private static final String LOGIN = "login";

	/** Logger for this class */
	private static final Logger logger = Logger.getLogger(XmlLoginModule.class.getName());

    private Set users;
	private Set globalPrincipals;
	private Set globalPrivateCredentials;
	private Set globalPublicCredentials;
	


    /**
     * initialize the loginModule.
     * @param subj
     * @param cbkHandler
     * @param sState
     * @param opts
     */
    public void initialize(Subject subj,CallbackHandler cbkHandler,Map sState,Map opts) {
		super.initialize(subj,cbkHandler,sState,opts);

        if (AuthenticationManagerFactory.getAuthenticationManager() == null){

        	Map newOpts = new HashMap();
    		newOpts.putAll(opts);

        	if (opts.get(CoreConstants.APPLICATION_NAME) != null){
        		// use XmlLoginModule options
        		newOpts.put(CoreConstants.APPLICATION_NAME, opts.get(CoreConstants.APPLICATION_NAME));

        	}else{
        		String appNameProp = System.getProperty("net.sf.jguard.application.name");

        		if (appNameProp != null){
        			// use system property net.sf.jguard.application.name
        			newOpts.put(CoreConstants.APPLICATION_NAME, appNameProp);

        		}else{
        			String appNameJMXProp = System.getProperty("com.sun.management.jmxremote.login.config");

        			if (appNameJMXProp != null){
        				logger.warning("Using JMX config for application name! " +
        						"If you're not running JMX, prefer XmlLoginModule options or net.sf.jguard.applicationName vmarg");
            			newOpts.put(CoreConstants.APPLICATION_NAME, appNameJMXProp);

        			}else{
        				//use default applicationName
            			newOpts.put(CoreConstants.APPLICATION_NAME, CoreConstants.DEFAULT_APPLICATION_NAME);
        			}
        		}
        	}

        	try {
        		AuthenticationManagerFactory.setAuthenticationManager(AuthenticationManagerFactory.createAuthenticationManager(XmlAuthenticationManager.class, newOpts));
			} catch (AuthenticationException e) {
				logger.log(Level.SEVERE, " initialize ", e);
			}
        }
 		try {
        	users = AuthenticationManagerFactory.getAuthenticationManager().getUsers();
        } catch (AuthenticationException e) {
			logger.log(Level.SEVERE, " initialize ", e);
		}
    }

    /**
	 * Authenticate the user.
	 * @return true if the user is authenticated, false otherwise.
	 * @exception FailedLoginException authentication fails
	 * @exception LoginException if this <code>LoginModule</code> is unable to perform the authentication.
	 */
	public boolean login() throws LoginException{
			super.login();

            JGuardCredential loginCredential = new JGuardCredential();
			loginCredential.setId(XmlLoginModule.LOGIN);
			loginCredential.setValue(login);

			JGuardCredential passwordCredential = new JGuardCredential();
			passwordCredential.setId(XmlLoginModule.PASSWORD);
			passwordCredential.setValue(new String(password));

	        Subject user;
	        Iterator it = users.iterator();
	        boolean authenticationSucceed = false;
	        
	        while(it.hasNext()){
	            user = (Subject)it.next();
	            Set privateCredentialsTemp = user.getPrivateCredentials();
	            if(privateCredentialsTemp.contains(loginCredential)){
	            	if((password!=null && privateCredentialsTemp.contains(passwordCredential))
	            	   ||skipPasswordCheck){

		            	//authentication succeed because one user has got cred1 and cred2
		            	globalPrincipals = user.getPrincipals();
		            	globalPrivateCredentials = user.getPrivateCredentials();
		            	globalPublicCredentials = user.getPublicCredentials();
		            	authenticationSucceed = true;
	            	}
	              break;
	            }
	        }

	        if(authenticationSucceed==false){
	        	loginOK = false;
	        	throw new FailedLoginException(XmlLoginModule.LOGIN_ERROR);
	        }

		return true;
	}


   	/**
	 * add Principals and Public/Private credentials to Subject.
	 * @see javax.security.auth.spi.LoginModule#commit()
	 */
	public boolean commit() throws LoginException {
		if(!loginOK){
			return false;
		}
		Set principals = subject.getPrincipals();
		if(globalPrincipals!=null){
		 principals.addAll(globalPrincipals);
		}
		Set privCredentials = subject.getPrivateCredentials();
		if(globalPrivateCredentials!=null){
		 privCredentials.addAll(globalPrivateCredentials);
		}
		Set pubCredentials = subject.getPublicCredentials();
		if(globalPublicCredentials!= null){
		 pubCredentials.addAll(globalPublicCredentials);
		}
		return true;
	}


}
