/*
 jGuard is a security framework based on top of jaas (java authentication and authorization security).
 it is written for web applications, to resolve simply, access control problems.
 version $Name$
 http://sourceforge.net/projects/jguard/

 Copyright (C) 2004  Charles GAY

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 jGuard project home page:
 http://sourceforge.net/projects/jguard/

 */
package net.sf.jguard.ext.authentication.manager;

import java.security.Principal;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.security.auth.Subject;
import javax.sql.DataSource;

import net.sf.jguard.core.CoreConstants;
import net.sf.jguard.core.authentication.AuthenticationException;
import net.sf.jguard.core.authentication.credentials.JGuardCredential;
import net.sf.jguard.core.authentication.manager.AuthenticationManager;
import net.sf.jguard.core.authentication.manager.AuthenticationManagerFactory;
import net.sf.jguard.core.principals.RolePrincipal;
import net.sf.jguard.core.provisioning.SubjectTemplate;
import net.sf.jguard.ext.JdbcManager;
import net.sf.jguard.ext.JdbcManagerHelper;
import net.sf.jguard.ext.SecurityConstants;
import net.sf.jguard.ext.database.ConnectionFactory;
import net.sf.jguard.ext.database.DatabaseUtils;
import net.sf.jguard.ext.principals.PrincipalUtils;

/**
 * AuthenticationManager implementation providing a database backend. differences between database "dialects", are handled through
 * specific properties files loaded during the init method.these properties files contains specific SQL queries.
 *
 * @author <a href="mailto:diabolo512@users.sourceforge.net">Charles Gay</a>
 * @author <a href="mailto:tandilero@users.sourceforge.net">Maximiliano Batelli</a>
 * @author <a href="mailto:vberetti@users.sourceforge.net">Vincent Beretti</a>
 * @see AuthenticationManager
 */
public class JdbcAuthenticationManager extends AbstractAuthenticationManager implements AuthenticationManager,JdbcManager {

	private static final Logger logger = Logger.getLogger(JdbcAuthenticationManager.class.getName());
	private Properties properties = null;
	
	private ConnectionFactory connectionFactory = null;

        /**
        * initialize this jdbc AuthorizationManager.
        * @param options a Map which contains informations to configure the AuthorizationManager implementation.
        */
        public JdbcAuthenticationManager(Map map) {
                super(map);
                // initialize database connection factory
		connectionFactory = new ConnectionFactory(map);
		init(map,connectionFactory);
	}
        
        /**
         * initialize this jdbc AuthenticationManager.
         * this constructor can be useful if you use one Inversion of Control (IoC) container.
         * this constructor can be used with Ioc type 3 (injection by constructor).
         * @param dataSource datasource to use to grab JDBC connections.
         * @param options a Map which contains informations to configure the AuthenticationManager implementation.
         */
        public JdbcAuthenticationManager(DataSource dataSource, Map map) {
                super(map);
                // initialize database connection factory
		connectionFactory = new ConnectionFactory(dataSource);
		init(map,connectionFactory);
	}
        
        private void init(Map options, ConnectionFactory connectionFactory){
        	Map jdbcMap = new HashMap();
        	String dbPropertiesLocation = (String)options.get(SecurityConstants.AUTHENTICATION_DATABASE_FILE_LOCATION);
            jdbcMap.put(JdbcManagerHelper.DB_PROPERTIES_LOCATION,dbPropertiesLocation);
            String createRequiredDatabaseEntities = (String)options.get(SecurityConstants.AUTHENTICATION_DATABASE_CREATE_REQUIRED_DATABASE_ENTITIES);
            jdbcMap.put(JdbcManagerHelper.CREATE_REQUIRED_DATABASE_ENTITIES, createRequiredDatabaseEntities);
            String importXmlDataKey = SecurityConstants.AUTHORIZATION_DATABASE_IMPORT_XML_DATA;
            jdbcMap.put(JdbcManagerHelper.IMPORT_XML_DATA_KEY, importXmlDataKey);
            String importXmlDataValue = (String)options.get(SecurityConstants.AUTHORIZATION_DATABASE_IMPORT_XML_DATA);
            jdbcMap.put(JdbcManagerHelper.IMP0RT_XML_DATA_VALUE, importXmlDataValue);
            jdbcMap.put(JdbcManagerHelper.XML_FILE_NAME, "jGuardUsersPrincipals.xml");
            properties = new Properties();
            JdbcManagerHelper.jdbcInit(this,connectionFactory,properties,jdbcMap);
		
		// we don't grab all the users like in XMLAuthenticationManager,
		// because XML is only used for small applications
		// it will be not suited to load all the users in memory

		try {
			initPrincipals();
		} catch (SQLException e) {
			logger.log(Level.SEVERE, " principals cannot be initialized ", e);
		}

		try {
			subjectTemplate = getSubjectTemplate(DEFAULT);
		} catch (AuthenticationException e) {
			logger.log(Level.SEVERE, " default subjectTemplate cannot be initialized ", e);
		}
            
        }
	protected void persistPrincipal(Principal principal) throws AuthenticationException {
		Connection connection = null;
		try {
			connection = connectionFactory.getConnection();
			PreparedStatement pst = connection.prepareStatement(properties.getProperty("ADD_PRINCIPAL"));
			
			if (principal instanceof RolePrincipal) {
				pst.setString(1, ((RolePrincipal) principal).getLocalName());
				pst.setString(2, ((RolePrincipal) principal).getApplicationName());
			} else {
				pst.setString(1, principal.getName());
				pst.setString(2, applicationName);
			}
			pst.setString(3, principal.getClass().getName());
			pst.executeUpdate();
			pst.close();

			localPrincipals.put(principal.getName(), principal);
			localPrincipalsSet.add(principal);

		} catch (SQLException e) {
			logger.log(Level.SEVERE, " persistPrincipal(Principal) ", e);
			throw new AuthenticationException(e);
		} finally {
			try {
				connection.close();
			} catch (SQLException e) {
				logger.log(Level.SEVERE, " persistPrincipal(Principal) ", e);
				throw new AuthenticationException(e);
			}
		}

	}

	

	public void createRequiredDatabaseEntities(Properties properties, ConnectionFactory connectionFactory) {
		List tablesNames = new ArrayList();
		tablesNames.add("JG_PRINCIPAL");
		tablesNames.add("JG_USER");
		tablesNames.add("JG_CREDENTIAL");
		tablesNames.add("JG_USER_PRINCIPAL");

		tablesNames.add("JG_ST_CREDENTIAL");
		tablesNames.add("JG_ST_PRINCIPAL");
		tablesNames.add("JG_ST_USER");
		
		tablesNames.add("JG_LOGON_AUDIT");
		
		List foreignkeysNames = new ArrayList();
		foreignkeysNames.add("FK_USER_USER_PRINCIPAL");
		foreignkeysNames.add("FK_USER_PRINCIPAL");
		foreignkeysNames.add("FK_CREDENTIAL_USER_ID");

		foreignkeysNames.add("FK_ST_USER_PRINCIPAL");
		foreignkeysNames.add("FK_ST_USER_CREDENTIAL");

		List sequencesNames = new ArrayList();
		sequencesNames.add("JG_PRINCIPAL_SEQ");
		sequencesNames.add("JG_USER_SEQ");
		sequencesNames.add("JG_CREDENTIAL_SEQ");

		sequencesNames.add("JG_ST_CREDENTIAL_SEQ");
		sequencesNames.add("JG_ST_PRINCIPAL_SEQ");
		sequencesNames.add("JG_ST_USER_SEQ");
		
		sequencesNames.add("JG_LOGON_AUDIT_SEQ");
		// initialise database
		DatabaseUtils.createRequiredDatabaseEntities(properties, connectionFactory, sequencesNames, tablesNames, foreignkeysNames);
	}

	/**
	 * initialize Principals declared in the application, and principals owned by the default
	 * @throws SQLException
	 * @see SubjectTemplate
	 */
	private void initPrincipals() throws SQLException {
		Connection connection = null;
		ResultSet rs = null;
		PreparedStatement pst = null;
		try{
		connection = connectionFactory.getConnection();
		pst = connection.prepareStatement(properties.getProperty("LOCAL_PRINCIPALS"));
		pst.setString(1, this.applicationName);
		rs = pst.executeQuery();
		while (rs.next()) {
			String name = rs.getString(2);
			String applicationName = rs.getString(3);
			String className = rs.getString(4);
			if (className.equals(RolePrincipal.class.getName())) {
				name = RolePrincipal.getName(name,applicationName);
			}
			Principal principal = PrincipalUtils.getPrincipal(className, name);
			localPrincipalsSet.add(principal);
			localPrincipals.put(principal.getName(), principal);
		}
		}finally{
			rs.close();
			pst.close();
			connection.close();
		}
	}

	public void insertRequiredData(String xmlFileLocation) {
		Map options = new HashMap();
		options.put(CoreConstants.APPLICATION_NAME, applicationName);
		options.put(SecurityConstants.AUTHENTICATION_XML_FILE_LOCATION, xmlFileLocation);
		
		try{
			AuthenticationManager authentManager = new XmlAuthenticationManager(options);
            importAuthenticationManager(authentManager);
		}catch(Exception e){
			logger.warning(e.getMessage());
			logger.warning(" XMLAuthenticationManager cannot be created with a jGuardUsersPrincipals.xml file to populate empty tables of your JdbcAuthenticationManager ");
			logger.warning(" database tables of your JDBCAuthenticationManager will stay empty ");
		}
		

	}

	public boolean isEmpty() {
		List selectQueries = new ArrayList();
		selectQueries.add("PRINCIPALS");
		selectQueries.add("CREDENTIALS");
		selectQueries.add("CREDENTIALS");
		selectQueries.add("USERS");

		return DatabaseUtils.isEmpty(this.properties, connectionFactory, selectQueries);

	}

	protected void updateUserImpl(JGuardCredential cred, Subject user) throws AuthenticationException {

		Connection connection = null;
		try {
			long subjectId = getSubjectId(cred, user);
			connection = connectionFactory.getConnection();
			// get user from DB
			Subject oldUser = getUser(subjectId);

			connection.setAutoCommit(false);

			updatePrivateCredentials(connection, subjectId, user, oldUser);
			updatePublicCredentials(connection, subjectId, user, oldUser);
			updatePrincipals(connection, subjectId, user, oldUser);

			connection.commit();
			connection.setAutoCommit(true);
		} catch (SQLException e) {
			logger.log(Level.SEVERE, " user can not be updated ", e);
			if (connection != null)
				try {
					connection.rollback();
				} catch (SQLException excep) {
					throw new AuthenticationException(e);
				}
			throw new AuthenticationException(e);
		} catch (SecurityException e) {
			logger.log(Level.FINEST, " user can not be updated: you don't have the permission to update private credentials", e);
			if (connection != null)
				try {
					connection.rollback();
				} catch (SQLException excep) {
					throw new AuthenticationException(excep);
				}
			throw new AuthenticationException(e);
		} finally {
			try {
				connection.close();
			} catch (SQLException e) {
				logger.log(Level.SEVERE, " user can not be updated ", e);
				throw new AuthenticationException(e);
			}
		}

	}

	//TODO think better how to deal with definition in RolePrincipal, because if we only update the definition attr
	// the RolePrincipal is not updated with this method
	// by now, we are deleting and creating again all user principals
	/*
	private void updatePrincipals(Connection connection, long subjectId, Subject user, Subject oldUser) throws SQLException {
		Set deletedPrincipals = new HashSet(oldUser.getPrincipals());
		deletedPrincipals.removeAll(user.getPrincipals());
		unlinkUserFromPrincipals(connection, subjectId, deletedPrincipals);

		Set newPrincipals = new HashSet(user.getPrincipals());
		newPrincipals.removeAll(oldUser.getPrincipals());
		linkUserToPrincipals(connection, subjectId, newPrincipals);
	}
	*/

	private void updatePrincipals(Connection connection, long subjectId, Subject user, Subject oldUser) throws SQLException {
		unlinkUserFromPrincipals(connection, subjectId, oldUser.getPrincipals());
		linkUserToPrincipals(connection, subjectId, user.getPrincipals());
	}
	
	private long getPrincipalId(Principal principal) throws SQLException {
		ResultSet rs = null;
		Connection connection = null;
		PreparedStatement pst = null;
		long principalId = -1;
		try{
		connection = connectionFactory.getConnection();
		pst = connection.prepareStatement(properties.getProperty("GET_PRINCIPAL_ID"));
		
		
		if (principal instanceof RolePrincipal) {
			pst.setString(1, ((RolePrincipal)principal).getLocalName());
			pst.setString(2, ((RolePrincipal) principal).getApplicationName());
		} else {
			pst.setString(1, principal.getName());
			pst.setString(2, applicationName);
		}
		rs = pst.executeQuery();
		rs.next();
		principalId = rs.getLong(1);
		}finally{
			rs.close();
			pst.close();
			connection.close();
		}
		
		return principalId;
	}

	private void unlinkUserFromPrincipals(Connection connection, long subjectId, Set principals) throws SQLException {

		PreparedStatement pst = connection.prepareStatement(properties.getProperty("UNLINK_USER_PRINCIPAL"));
		Iterator it = principals.iterator();
		while (it.hasNext()) {
			Principal principal = (Principal) it.next();
			long principalId = getPrincipalId(principal);
			pst.setLong(1, subjectId);
			pst.setLong(2, principalId);
			pst.executeUpdate();
		}
		pst.close();
	}

	private void linkUserToPrincipals(Connection connection, long subjectId, Set principals) throws SQLException {
		PreparedStatement pst = connection.prepareStatement(properties.getProperty("LINK_USER_PRINCIPAL"));
		Iterator it = principals.iterator();
		while (it.hasNext()) {
			Principal principal = (Principal) it.next();
			long principalId = getPrincipalId(principal);
			pst.setLong(1, subjectId);
			pst.setLong(2, principalId);
			if(principal instanceof RolePrincipal) {
				pst.setString(3, ((RolePrincipal)principal).getDefinition());
				pst.setBoolean(4, ((RolePrincipal)principal).isActive());
			}
			pst.executeUpdate();
		}
		pst.close();
	}

	private void updatePrivateCredentials(Connection connection, long subjectId, Subject user, Subject oldUser)
			throws SQLException {

		// delete old private credentials
		Set deletedPrivateCredentials = new HashSet(oldUser.getPrivateCredentials(JGuardCredential.class));
		deletedPrivateCredentials.removeAll(user.getPrivateCredentials(JGuardCredential.class));
		Iterator itDeletedPrivateCredentials = deletedPrivateCredentials.iterator();
		while (itDeletedPrivateCredentials.hasNext()) {
			deleteCredential(connection, (JGuardCredential) itDeletedPrivateCredentials.next(), subjectId);
		}

		// add new private credentials
		Set newPrivateCredentials = new HashSet(user.getPrivateCredentials(JGuardCredential.class));
		newPrivateCredentials.removeAll(oldUser.getPrivateCredentials(JGuardCredential.class));
		Iterator itNewPrivateCredentials = newPrivateCredentials.iterator();
		while (itNewPrivateCredentials.hasNext()) {
			persistCredential(connection, (JGuardCredential) itNewPrivateCredentials.next(), subjectId, false);
		}
	}

	private void updatePublicCredentials(Connection connection, long subjectId, Subject user, Subject oldUser)
			throws SQLException {

		// delete old public credentials
		Set deletedPublicCredentials = new HashSet(oldUser.getPublicCredentials(JGuardCredential.class));
		deletedPublicCredentials.removeAll(user.getPublicCredentials(JGuardCredential.class));
		Iterator itDeletedPublicCredentials = deletedPublicCredentials.iterator();
		while (itDeletedPublicCredentials.hasNext()) {
			deleteCredential(connection, (JGuardCredential) itDeletedPublicCredentials.next(), subjectId);
		}

		// add new public credentials
		Set newPublicCredentials = new HashSet(user.getPublicCredentials(JGuardCredential.class));
		newPublicCredentials.removeAll(oldUser.getPublicCredentials(JGuardCredential.class));
		Iterator itNewPublicCredentials = newPublicCredentials.iterator();
		while (itNewPublicCredentials.hasNext()) {
			persistCredential(connection, (JGuardCredential) itNewPublicCredentials.next(), subjectId, true);
		}
	}

	public void deleteUser(Subject user) throws AuthenticationException {
		
		PreparedStatement pst = null;
		Connection connection = null;
		try {
			connection = connectionFactory.getConnection();
			connection.setAutoCommit(false);
			long subjectId = getSubjectId(user);

			deleteCredentials(connection, subjectId);
			deletePrincipals(connection, subjectId);

			pst = connection.prepareStatement(properties.getProperty("DELETE_SUBJECT"));
			pst.setLong(1, subjectId);
			pst.executeUpdate();
			pst.close();

			connection.commit();
			connection.setAutoCommit(true);

		} catch (SQLException e) {
			logger.log(Level.SEVERE, "deleteUser(Subject user)", e);
			if (connection != null)
				try {
					connection.rollback();
				} catch (SQLException excep) {
					throw new AuthenticationException(excep);
				}
			throw new AuthenticationException(e);
		} finally {
			try {
				connection.close();
			} catch (SQLException e) {
				logger.log(Level.SEVERE, "deleteUser(Subject user)", e);
				throw new AuthenticationException(e);
			}
		}

	}

	private long getSubjectId(Subject user) throws SQLException, AuthenticationException {
		JGuardCredential identityCred = extractIdentityCredentialFromUser(user);
		return getSubjectId(identityCred, user);
	}

	private long getSubjectId(JGuardCredential identityCred, Subject user) throws SQLException, AuthenticationException {
		long userId = -1;
		Connection connection = null;
		PreparedStatement pst = null;
		ResultSet rs = null;
		try{
		connection = connectionFactory.getConnection();
		pst = connection.prepareStatement(properties.getProperty("GET_USER_ID"));
		pst.setString(1, identityCred.getId());
		pst.setString(2, identityCred.getValue().toString());

		rs = pst.executeQuery();
		while (rs.next()) {
			if (userId != -1) {
				logger.severe("it exists more than one user with this identity credential ");
				throw new AuthenticationException("it exists more than one user with this identity credential ");
			}
			userId = rs.getLong(1);
		}
		if (userId == -1) {
			logger.severe("no user exists with this identity credential ");
			throw new AuthenticationException("no user exists with this identity credential ");
		}
		}finally{
			rs.close();
			pst.close();
			connection.close();
		}
		return userId;
	}

	private void deletePrincipals(Connection connection, long subjectId) throws SQLException {
		PreparedStatement pst = null;
		pst = connection.prepareStatement(properties.getProperty("DELETE_USER_PRINCIPALS"));
		pst.setLong(1, subjectId);
		pst.executeUpdate();
		pst.close();
	}

	private void deleteCredentials(Connection connection, long subjectId) throws SQLException {
		PreparedStatement pst = null;
		pst = connection.prepareStatement(properties.getProperty("DELETE_USER_CREDENTIALS"));
		pst.setLong(1, subjectId);
		pst.executeUpdate();
		pst.close();
	}

	private void persistCredentials(Connection connection, Subject subject, long subjectId) throws SQLException {
		Set publicCredentials = subject.getPublicCredentials();
		Iterator itPublic = publicCredentials.iterator();
		while (itPublic.hasNext()) {
			JGuardCredential jgCred = (JGuardCredential) itPublic.next();
			persistCredential(connection, jgCred, subjectId, true);
		}

		Set privateCredentials = subject.getPrivateCredentials();
		Iterator itPrivate = privateCredentials.iterator();
		while (itPrivate.hasNext()) {
			JGuardCredential jgCred = (JGuardCredential) itPrivate.next();
			persistCredential(connection, jgCred, subjectId, false);
		}

	}

	private void persistCredential(Connection connection, JGuardCredential jgCred, long userId, boolean publicVisibility)
			throws SQLException {

		PreparedStatement pst = connection.prepareStatement(properties.getProperty("ADD_CREDENTIAL"));
		pst.setLong(1, userId);
		pst.setBoolean(2, publicVisibility);
		pst.setString(3, jgCred.getId());
		pst.setString(4, jgCred.getValue().toString());
		pst.executeUpdate();
		pst.close();
	}

	// TODO VBE : remove this method. Credentials are not updated anymore, they are removed and added.
	// This is done because this method is not handling multivalued credentials
	private void updateCredential(Connection connection, JGuardCredential jgCred, long userId) throws SQLException {

		PreparedStatement pst = connection.prepareStatement(properties.getProperty("UPDATE_CREDENTIAL"));
		pst.setString(1, jgCred.getValue().toString());
		pst.setLong(2, userId);
		pst.setString(3, jgCred.getId());
		pst.executeUpdate();
		pst.close();
	}

	private void deleteCredential(Connection connection, JGuardCredential jgCred, long userId) throws SQLException {

		PreparedStatement pst = connection.prepareStatement(properties.getProperty("DELETE_CREDENTIAL"));
		pst.setLong(1, userId);
		pst.setString(2, jgCred.getId());
		pst.setString(3, (jgCred.getValue() != null ? jgCred.getValue().toString() : ""));
		pst.executeUpdate();
		pst.close();
	}

	private void persistPrincipals(Connection connection, Subject subject, long subjectId) throws SQLException,
			AuthenticationException {

		PreparedStatement pst = connection.prepareStatement(properties.getProperty("LINK_USER_PRINCIPAL"));
		Set principals = subject.getPrincipals();
		Iterator it = principals.iterator();
		while (it.hasNext()) {
			Principal ppal = (Principal) it.next();
			createPrincipal(ppal);

			long principalId = getPrincipalId(ppal);

			pst.setLong(1, subjectId);
			pst.setLong(2, principalId);
			if(ppal instanceof RolePrincipal) {
				pst.setString(3, ((RolePrincipal)ppal).getDefinition());
				pst.setBoolean(4, ((RolePrincipal)ppal).isActive());
			}
			
			pst.executeUpdate();
		}
		pst.close();
	}

	protected void persistUser(Subject user) throws AuthenticationException {
		
		PreparedStatement pst = null;
		PreparedStatement pst2 = null;
		ResultSet rs2 = null;
		Connection connection = null;
		try {
			connection = connectionFactory.getConnection();
			connection.setAutoCommit(false);

			pst = connection.prepareStatement(properties.getProperty("ADD_SUBJECT"));
			pst.executeUpdate();
			pst.close();

			pst2 = connection.prepareStatement(properties.getProperty("GET_SUBJECT_ID"));
			rs2 = pst2.executeQuery();
			rs2.next();
			long subjectId = rs2.getLong(1);
			rs2.close();
			pst2.close();

			persistCredentials(connection, user, subjectId);
			persistPrincipals(connection, user, subjectId);

			connection.commit();
			connection.setAutoCommit(true);

		} catch (SQLException e) {
			logger.log(Level.SEVERE, "persistUser(Subject user)", e);
			if (connection != null)
				try {
					connection.rollback();
				} catch (SQLException excep) {
					throw new AuthenticationException(excep);
				}
			throw new AuthenticationException(e);
		} finally {
			try {
				connection.close();
			} catch (SQLException e) {
				logger.log(Level.SEVERE, "persistUser(Subject user)", e);
				throw new AuthenticationException(e);
			}
		}

	}

	public void updateDefaultSubjectTemplate(SubjectTemplate template) throws AuthenticationException {
		Connection connection = null;

		try {
			connection = connectionFactory.getConnection();
			connection.setAutoCommit(false);
			long defaultSubjecttemplateID = getSubjectTemplateID(DEFAULT);
			persistSubjectTemplateCredentials(connection, template, defaultSubjecttemplateID);
			persistSubjectTemplatePrincipals(connection, template, defaultSubjecttemplateID);

			connection.commit();
			connection.setAutoCommit(true);
		} catch (SQLException e) {
			logger.log(Level.SEVERE, " default subject_template cannot be updated ", e);
			if (connection != null)
				try {
					connection.rollback();
				} catch (SQLException excep) {
					throw new AuthenticationException(excep);
				}
			throw new AuthenticationException(e);
		}finally {
			try {
				connection.close();
			} catch (SQLException e) {
				logger.log(Level.SEVERE, "default subject_template cannot be updated ", e);
				throw new AuthenticationException(e);
			}
		}

	}

	public void persistSubjectTemplate(SubjectTemplate template) throws AuthenticationException {
		Connection connection = null;
		long stId = -1;
		try {
			connection = connectionFactory.getConnection();
			connection.setAutoCommit(false);

			PreparedStatement pst = connection.prepareStatement(properties.getProperty("ADD_SUBJECT_TEMPLATE"));
			pst.setString(1, template.getName());
			pst.executeUpdate();
			pst.close();

			PreparedStatement pst2 = connection.prepareStatement(properties.getProperty("GET_SUBJECT_TEMPLATE_ID"));
			ResultSet rs2 = pst2.executeQuery();
			rs2.next();
			stId = rs2.getLong(1);
			rs2.close();
			pst2.close();

			persistSubjectTemplateCredentials(connection, template, stId);
			persistSubjectTemplatePrincipals(connection, template, stId);

			connection.commit();
			connection.setAutoCommit(true);
		} catch (SQLException e) {
			logger.log(Level.SEVERE, " subject_template cannot be created ", e);
			if (connection != null)
				try {
					connection.rollback();
				} catch (SQLException excep) {
					throw new AuthenticationException(excep);
				}
			throw new AuthenticationException(e);
		} finally {
			try {
				connection.close();
			} catch (SQLException e) {
				logger.log(Level.SEVERE, "subject_template cannot be created", e);
				throw new AuthenticationException(e);
			}
		}
	}

	private void persistSubjectTemplatePrincipals(Connection connection, SubjectTemplate template, long stId) throws SQLException {

		Set genericPrincipals = template.getPrincipals();
		Iterator itGenPrincipals = genericPrincipals.iterator();
		while (itGenPrincipals.hasNext()) {
			Principal ppal = (Principal) itGenPrincipals.next();
			persistSubjectTemplatePrincipal(connection, ppal, stId);
		}
	}

	private void persistSubjectTemplatePrincipal(Connection connection, Principal principal, long stId) throws SQLException {

		PreparedStatement pst = connection.prepareStatement(properties.getProperty("ADD_ST_PRINCIPAL"));
		if (principal instanceof RolePrincipal) {
			pst.setString(1, ((RolePrincipal)principal).getLocalName());
			pst.setString(2, ((RolePrincipal) principal).getApplicationName());
		} else {
			pst.setString(1, principal.getName());
			pst.setString(2, applicationName);
		}
		pst.setString(3, principal.getClass().getName());

		pst.setLong(4, stId);
		pst.executeUpdate();
		pst.close();

		localPrincipals.put(principal.getName(), principal);
		localPrincipalsSet.add(principal);
	}

	private void persistSubjectTemplateCredentials(Connection connection, SubjectTemplate st, long subjectTemplateId)
			throws SQLException {

		Set privateRequiredCred = st.getPrivateRequiredCredentials();
		Iterator itPrivateRequiredCred = privateRequiredCred.iterator();
		while (itPrivateRequiredCred.hasNext()) {
			JGuardCredential jgCred = (JGuardCredential) itPrivateRequiredCred.next();
			// required and public: 'true' and 'false'
			persistStCredential(connection, jgCred, subjectTemplateId, true, false);
		}

		Set publicRequiredCreds = st.getPublicRequiredCredentials();
		Iterator itPublicRequiredCred = publicRequiredCreds.iterator();
		while (itPublicRequiredCred.hasNext()) {
			JGuardCredential jgCred = (JGuardCredential) itPublicRequiredCred.next();
			// required and public: both 'true'
			persistStCredential(connection, jgCred, subjectTemplateId, true, true);
		}

		Set privateOptionalCred = st.getPrivateOptionalCredentials();
		Iterator itPrivateOptionalCred = privateOptionalCred.iterator();
		while (itPrivateOptionalCred.hasNext()) {
			JGuardCredential jgCred = (JGuardCredential) itPrivateOptionalCred.next();
			// required and public: both 'false'
			persistStCredential(connection, jgCred, subjectTemplateId, false, false);
		}

		Set publicOptionalCreds = st.getPublicOptionalCredentials();
		Iterator itPublicOptionalCred = publicOptionalCreds.iterator();
		while (itPublicOptionalCred.hasNext()) {
			JGuardCredential jgCred = (JGuardCredential) itPublicOptionalCred.next();
			// required and public: 'false' and 'true'
			persistStCredential(connection, jgCred, subjectTemplateId, false, true);
		}
	}

	private void persistStCredential(Connection connection, JGuardCredential jgCred, long subjectTemplateId, boolean required,
			boolean public_visibility) throws SQLException {

		PreparedStatement pst = connection.prepareStatement(properties.getProperty("ADD_ST_CREDENTIAL"));
		pst.setString(1, jgCred.getId());
		pst.setBoolean(2, public_visibility);
		pst.setBoolean(3, required);
		pst.setBoolean(4, jgCred.isIdentity());
		pst.setLong(5, subjectTemplateId);
		pst.executeUpdate();
		pst.close();
	}

	public final ConnectionFactory getConnectionFactory() {
		return connectionFactory;
	}

	public final Properties getProperties() {
		return properties;
	}

	public final void setProperties(Properties properties) {
		this.properties = properties;
	}

	public SubjectTemplate getSubjectTemplate(String name) throws AuthenticationException {
		
		SubjectTemplate st = new SubjectTemplate();
		try {
			long subjectTemplateId = getSubjectTemplateID(name);
			populateSubjectTemplatePrincipals(st, subjectTemplateId);
			populateSubjectTemplateCredentials(st, subjectTemplateId);

		} catch (SQLException e) {
			logger.log(Level.SEVERE, " subject_template cannot be grabbed ", e);
			throw new AuthenticationException(e);
		}

		return st;

	}

	private void populateSubjectTemplatePrincipals(SubjectTemplate st, long subjectTemplateId) throws SQLException {
		Connection connection= null;
		
		try{
			connection = connectionFactory.getConnection();
			PreparedStatement pst2 = connection.prepareStatement(properties.getProperty("GET_ST_PRINCIPALS"));
			pst2.setLong(1, subjectTemplateId);
			ResultSet rs2 = pst2.executeQuery();
			Set genericPrincipals = new HashSet();
			while (rs2.next()) {
				String className = rs2.getString(4);
				String principalName = (String) rs2.getString(2);
				if (className.equals(RolePrincipal.class.getName())) {
					principalName = RolePrincipal.getName(principalName,rs2.getString(3));
				}
				Principal ppal = PrincipalUtils.getPrincipal(className, principalName);
				genericPrincipals.add(ppal);
			}
			rs2.close();
			pst2.close();
			st.setPrincipals(genericPrincipals);
		}finally{
			connection.close();
		}
		
	}

	private void populateSubjectTemplateCredentials(SubjectTemplate st, long subjectTemplateId) throws SQLException {
		Connection connection = null;
		try{
			
		connection = connectionFactory.getConnection();
		PreparedStatement pst3 = connection.prepareStatement(properties.getProperty("GET_ST_CREDENTIALS"));
		pst3.setLong(1, subjectTemplateId);
		ResultSet rs3 = pst3.executeQuery();
		Set privateOptionalCredentials = new HashSet();
		Set privateRequiredCredentials = new HashSet();
		Set publicOptionalCredentials = new HashSet();
		Set publicRequiredCredentials = new HashSet();

		while (rs3.next()) {
			JGuardCredential cred = new JGuardCredential();
			cred.setId(rs3.getString(2));
			boolean public_visibility = rs3.getBoolean(3);
			boolean required = rs3.getBoolean(4);
			cred.setIdentity(rs3.getBoolean(5));
			if (public_visibility && required) {
				publicRequiredCredentials.add(cred);
			} else if (!public_visibility && required) {
				privateRequiredCredentials.add(cred);
			} else if (public_visibility && !required) {
				publicOptionalCredentials.add(cred);
			} else if (!public_visibility && !required) {
				privateOptionalCredentials.add(cred);
			}

		}
		rs3.close();
		pst3.close();
		st.setPrivateOptionalCredentials(privateOptionalCredentials);
		st.setPrivateRequiredCredentials(privateRequiredCredentials);
		st.setPublicOptionalCredentials(publicOptionalCredentials);
		st.setPublicRequiredCredentials(publicRequiredCredentials);
		}finally{
			connection.close();
		}
	}

	/**
	 * return the row id for the {@link SubjectTemplate} identified by its name.
	 * @param name
	 * @return
	 * @throws SQLException
	 * @throws AuthenticationException thrown when either no ID is found or more than one ID are found
	 */
	private long getSubjectTemplateID(String name) throws SQLException, AuthenticationException {
		Connection connection = null;
		PreparedStatement pst;
		ResultSet rs;
		long subjectTemplateId = -1;
		int subjectTemplateFound = 0;
		try{
			connection = connectionFactory.getConnection();
			pst = connection.prepareStatement(properties.getProperty("GET_SUBJECT_TEMPLATE"));
			pst.setString(1, name);
			rs = pst.executeQuery();
			
			while (rs.next()) {
				subjectTemplateId = rs.getLong(1);
				subjectTemplateFound++;
			}
			rs.close();
			pst.close();
			
		}finally{
			connection.close();
		}
		if(subjectTemplateId==-1){
			throw new AuthenticationException(" subjectTemplate ID intitled "+name+" cannot be found ");
		}
		if(subjectTemplateFound>1){
			throw new AuthenticationException(" more than one ID  for subjectTemplate intitled "+name+" has been found ");
		}
		return subjectTemplateId;
	}

	public Set getAllPrincipalsSet() throws AuthenticationException {
		Set allPrincipals = new HashSet();
		
		PreparedStatement pst = null;
		ResultSet rs = null;
		Connection connection = null;
		try {
			connection = connectionFactory.getConnection();
			pst = connection.prepareStatement(properties.getProperty("PRINCIPALS"));
			rs = pst.executeQuery();
			while (rs.next()) {
				String name = rs.getString(2);
				String className = rs.getString(4);
				String applicationName = rs.getString(3);
				if (RolePrincipal.class.getName().equals(className)) {
					name=  RolePrincipal.getName(name, applicationName);
				}
				Principal principal = PrincipalUtils.getPrincipal(className, name);
				
				allPrincipals.add(principal);
			}

			rs.close();
			pst.close();
		} catch (SQLException e) {
			logger.log(Level.SEVERE, " principals cannot be grabbed ", e);
			throw new AuthenticationException(e);
		} finally {
			try {
				connection.close();
			} catch (SQLException e) {
				logger.log(Level.SEVERE, " principals cannot be grabbed ", e);
				throw new AuthenticationException(e);
			}
		}
		return allPrincipals;
	}

	/**
	 * search the users which matches ALL of the credentials criterions.
	 *
	 * @param credentials
	 *            crierions used to grab the users
	 * @return users found
	 */
	public Set findUsers(Collection credentials) throws AuthenticationException {
		Set users = new HashSet();

		if (credentials.isEmpty())
			return users;

		Connection connection = null;
		PreparedStatement pst = null;
		ResultSet rs = null;
		String findUsers = properties.getProperty("FIND_USERS");
		String[] tokens = findUsers.split(" ");
		String user_id = tokens[2];
		String jg_credential = tokens[4];
		String cred_name = tokens[6];
		String cred_value = tokens[9];
		try {
			connection = connectionFactory.getConnection();
			// we build the query
			StringBuffer dynamicQuery = new StringBuffer();
			StringBuffer tables = new StringBuffer();
			StringBuffer conditions = new StringBuffer();

			for (int count = 0; count < credentials.size(); count++) {
				if (count > 0) {
					tables.append(',');
					conditions.append(" and ");
				} else {
					conditions.append(" where ");
				}
				StringBuffer alias = new StringBuffer("cred").append(count);
				tables.append(' ').append(jg_credential).append(' ').append(alias);
				conditions.append(alias).append('.').append(cred_name).append("= ? and ");
				conditions.append(alias).append('.').append(cred_value).append("= ? ");
				if (count > 0) {
					conditions.append(" and cred0.").append(user_id).append('=');
					conditions.append(alias).append('.').append(user_id).append(' ');
				}
			}

			dynamicQuery.append(" select distinct cred0.").append(user_id).append(" from ");
			dynamicQuery.append(tables);
			dynamicQuery.append(conditions);
			logger.finest(dynamicQuery.toString());
			pst = connection.prepareStatement(dynamicQuery.toString());

			Iterator itCredentials = credentials.iterator();
			int count2 = 1;
			while (itCredentials.hasNext()) {
				JGuardCredential jcred = (JGuardCredential) itCredentials.next();
				pst.setString(count2, jcred.getId());
				count2++;
				pst.setString(count2, jcred.getValue().toString());
				count2++;
			}
			rs = pst.executeQuery();
			while (rs.next()) {
				Subject user = null;
				int userId = rs.getInt(1);
				user = getUser(userId);
				users.add(user);
			}

			rs.close();
			pst.close();
		} catch (SQLException e) {
			logger.log(Level.SEVERE, " error when we try to find users which maps to credentials ", e);
			throw new AuthenticationException(e);
		} finally {
			try {
				connection.close();
			} catch (SQLException e) {
				logger.log(Level.SEVERE, "users cannot be found ", e);
				throw new AuthenticationException(e);
			}
		}
		return users;
	}

	/**
	 * search the user's principals .
	 *
	 * @param usrId
	 * @return user principals
	 * @throws SQLException
	 */
	private static Set getUserPrincipals(long usrId) throws SQLException {
		JdbcAuthenticationManager authManager = (JdbcAuthenticationManager) AuthenticationManagerFactory
				.getAuthenticationManager();
		ConnectionFactory connectionFactory = authManager.getConnectionFactory();
		Connection connection = null;
		PreparedStatement pst = null;
		ResultSet rs = null;
		Set userPrincipals = new HashSet();
		try{
			connection = connectionFactory.getConnection();
			
			pst = connection.prepareStatement(authManager.getProperties().getProperty("GET_USER_PRINCIPALS"));
			pst.setLong(1, usrId);
			rs = pst.executeQuery();
			while (rs.next()) {
				RolePrincipal principal = new RolePrincipal();
				principal.setLocalName(rs.getString(1));
				principal.setApplicationName(rs.getString(2));
				logger.finest("principal added" + principal + " to user " + usrId);
				userPrincipals.add(principal);
				
			}
		}catch(SQLException e){
			logger.warning(" principals for user with id="+usrId+" cannot be found");
			throw e;
			
		}finally{
		rs.close();
		pst.close();
		connection.close();
		}
		return userPrincipals;
	}

	/**
	 * search the user's credentials .
	 *
	 * @param usrId
	 * @param publicVisibility - indicate if the credentials to return are public or private
	 * @return user credentials
	 * @throws SQLException
	 */
	private static Set getUserCredentials(long usrId, boolean publicVisibility) throws SQLException {
		JdbcAuthenticationManager authManager = (JdbcAuthenticationManager) AuthenticationManagerFactory
				.getAuthenticationManager();
		ConnectionFactory connectionFactory = authManager.getConnectionFactory();
		Connection connection = null;
		PreparedStatement pst = null;
		ResultSet rs = null;
		Set userCredentials = new HashSet();
		try{
		connection = connectionFactory.getConnection();
		
		pst = connection.prepareStatement(authManager.getProperties().getProperty("GET_USER_CREDENTIALS"));
		pst.setLong(1, usrId);
		pst.setBoolean(2, publicVisibility);
		rs = pst.executeQuery();
		while (rs.next()) {
			JGuardCredential cred = new JGuardCredential();
			cred.setId(rs.getString(1));
			cred.setValue(rs.getString(2));
			userCredentials.add(cred);
			logger.finest(" credential added " + cred + " to user " + usrId);
		}
		}finally{
			rs.close();
			pst.close();
			connection.close();
		}
		
		return userCredentials;
	}

	/**
	 * return all users of jGuard.
	 */
	public Set getUsers() throws AuthenticationException {
		Set users = new HashSet();
		
		PreparedStatement pst = null;
		ResultSet rs = null;
		Connection connection = null;
		try {
			connection = connectionFactory.getConnection();
			pst = connection.prepareStatement(properties.getProperty("GET_USERS"));
			rs = pst.executeQuery();
			while (rs.next()) {
				long userId = rs.getLong(1);
				Subject user = getUser(userId);
				users.add(user);
			}

			rs.close();
			pst.close();
		} catch (SQLException e) {
			logger.log(Level.SEVERE, " users cannot be grabbed ", e);
			throw new AuthenticationException(e);
		} finally {
			try {
				connection.close();
			} catch (SQLException e) {
				logger.log(Level.SEVERE, " users cannot be grabbed ", e);
				throw new AuthenticationException(e);
			}
		}
		return users;
	}

	/**
	 *
	 * @param userId
	 * @return 
	 * @throws SQLException
	 */
	public static Subject getUser(long userId) throws SQLException {
		// SQL exception is not catched because JdbcLoginModule need to have the exception to rollback
		Set userPrincipals = getUserPrincipals(userId);
		Set userPubCred = getUserCredentials(userId, true);
		Set userPrivCred = getUserCredentials(userId, false);
		// subject is not in read-only mode
		Subject user = new Subject(false, userPrincipals, userPubCred, userPrivCred);
		return user;
	}

	public void updatePrincipal(String oldName, Principal principal) throws AuthenticationException {
		if (oldName.equals(principal.getName())) {
			return;
		}
		Connection connection = null;
		PreparedStatement pst = null;
		try {
			connection = connectionFactory.getConnection();
			pst = connection.prepareStatement(properties.getProperty("UPDATE_PRINCIPAL"));
			pst.setString(1, principal.getName());
			pst.setString(2, oldName);
			pst.setString(3, applicationName);
			pst.executeUpdate();
			pst.close();

		} catch (SQLException e) {
			logger.log(Level.SEVERE, "updatePrincipal(String oldName, Principal principal)", e);
			throw new AuthenticationException(e);
		} finally {
			try {
				connection.close();
			} catch (SQLException e) {
				logger.log(Level.SEVERE, "updatePrincipal(String oldName, Principal principal)", e);
				throw new AuthenticationException(e);
			}
		}

		// if the database update is right, we can update the in memory reference
		Principal oldPal = (Principal)localPrincipals.remove(oldName);
		localPrincipalsSet.remove(oldPal);
		localPrincipals.put(principal.getName(), principal);
		localPrincipalsSet.add(principal);
		

	}

	public boolean deletePrincipal(Principal principal) throws AuthenticationException {
		
		PreparedStatement pst = null;
		PreparedStatement pst2 = null;
		PreparedStatement pst3 = null;
		Connection connection = null;
		ResultSet rs = null;
		try {
			connection = connectionFactory.getConnection();
			connection.setAutoCommit(false);

			// we are looking for the principalId
			pst = connection.prepareStatement(properties.getProperty("GET_PRINCIPAL_ID"));
			pst.setString(1, principal.getName());
			pst.setString(2, applicationName);
			rs = pst.executeQuery();
			long principalId = -1;
			if (rs.next()){
				principalId = rs.getLong(1);
			}else{
				logger.finest(" there is no principal to delete: " + principal.getName());
				return false;
			}
			

			// we use principalId to delete links between users and this principal
			pst2 = connection.prepareStatement(properties.getProperty("DELETE_PRINCIPAL_USERS"));
			pst2.setLong(1, principalId);
			pst2.executeUpdate();
			pst2.close();

			// we delete the principal
			pst3 = connection.prepareStatement(properties.getProperty("DELETE_PRINCIPAL"));
			pst3.setString(1, principal.getName());
			pst3.executeUpdate();
			pst3.close();

			connection.commit();
			connection.setAutoCommit(true);
			
		} catch (SQLException e) {
			logger.log(Level.SEVERE, "deletePrincipal(Principal principal)", e);
			if (connection != null)
				try {
					connection.rollback();
				} catch (SQLException excep) {
					throw new AuthenticationException(e);
				}
			throw new AuthenticationException(e);
		} finally {
			try {
				rs.close();
				pst.close();
				connection.close();
			} catch (SQLException e) {
				logger.log(Level.SEVERE, "deletePrincipal(Principal principal)", e);
				throw new AuthenticationException(e);
			}
		}
		
		// if the database update is right, we can update the in memory reference
		Principal oldPal = (Principal)localPrincipals.remove(principal.getName());
		localPrincipalsSet.remove(oldPal);
		return true;
	}
        
        /**
	 * Persists a user logon attempt.
         * @param login user unique ID
         * @param hostAdress IP Adress
         * @param hostname DNS Name bound to the IP adress
         * @succeeded boolean authentication result
         * @param timestamp long tracing the timestamp when the user has tried to logon
	 */
	public void persistUserAccessAttempt(String login, String hostAddress, String hostName, boolean succeeded,long timeStamp) throws AuthenticationException {
                PreparedStatement pst = null;
                Connection connection = null;
                try{
                    connection = connectionFactory.getConnection();
                    pst = connection.prepareStatement(properties.getProperty("ADD_LOGON_AUDIT"));
                    pst.setString(1, login);
                    pst.setString(2, hostAddress);
                    pst.setString(3, hostName);
                    pst.setBoolean(4, succeeded);
                    pst.setTimestamp(5,new Timestamp(timeStamp));
                    pst.executeUpdate();
                    pst.close();		
                }catch(SQLException sqle){
                    throw new AuthenticationException(sqle);
                }finally{
                    try {
                        connection.close();
                    } catch (SQLException ex) {
                        throw new AuthenticationException(ex);
                    }
                }
	}

    
	
}
